#define	HERE_CHARYBDIS

#include "util.h"

#include <netinet/tcp.h>
#include <time.h>

static long	totalcount;
static char	workmode;

#define	INTERPOLATION 30

static void
alarm_ticker(void)
{
  static int	speedometer[INTERPOLATION];
  static time_t	timing[INTERPOLATION];
  static int	lastcount;
  static long	delta;
  static int	pos;
  time_t	now;
  long		seconds;
  double	bytes, percent;

  pos			= (pos+1)%INTERPOLATION;

  time(&now);

  delta			-= speedometer[pos];
  delta			+= 
    speedometer[pos]	= copycount_in-lastcount;
  lastcount		= copycount_in;

  seconds		= now-timing[pos];
  timing[pos]		= now;

  bytes			= (double)delta/seconds/1024.;
  percent		= -1;
  if (totalcount)
    percent		= copycount_in*100./totalcount;

  printf("\r%c %7ld %6.2f%% %8.2f %2d ",
	 workmode, (totalcount-copycount_in)>>10,
	 percent, bytes, alarm_timeout);
  fflush(stdout);
}

static void
getaddr(struct sockaddr_in *sin, const char *host, int port)
{
  memset(sin, 0, sizeof *sin);
  sin->sin_family	= AF_INET;
  sin->sin_addr.s_addr	= htonl(INADDR_ANY);
  sin->sin_port		= htons(port);
  if (host && !inet_aton(host, &sin->sin_addr))
    {
      struct hostent	*he;

      if ((he=gethostbyname(host))==0)
	ex(host);
      if (he->h_addrtype!=AF_INET || he->h_length!=sizeof sin->sin_addr)
	ex("unsupported host address");
      memcpy(&sin->sin_addr, he->h_addr, sizeof sin->sin_addr);
    }
}

static void
sockopen(const char *host, int port)
{
  struct sockaddr_in	sin;
  int			on;
  struct linger		linger;
  unsigned int		mss;

  sock	= socket( AF_INET, SOCK_STREAM, 0 );
  if (sock<0)
    ex("socket");

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on))
    ex("setsockopt reuse");
  linger.l_onoff	= 1;
  linger.l_linger	= 65535;
  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger))
    ex("setsockopt linger");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &on, sizeof on))
    ex("setsockopt sndbuf");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &on, sizeof on))
    ex("setsockopt rcvbuf");

  /* 2002-05-09 set the MTU
   * The kernel takes a lower value if it is not supported ;)
   * We add some bytes for the tcp packet overhead.
   */
  if (maxblock<1500)
    {
      mss = maxblock+48;
      if(setsockopt(sock, IPPROTO_TCP, TCP_MAXSEG, &mss, sizeof(mss)))
        ex("setsockopt maxseg");
    }

  getaddr(&sin, host, port);

  if (bind(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("bind");

#if 0
  if (fcntl(sock, F_SETFL, O_NDELAY))
    ex("O_NDELAY");
#endif
}

static void
do_connect(const char *host, int port)
{
  struct sockaddr_in	sin;

  sockopen(NULL, 0);

  getaddr(&sin, host, port);
  if (connect(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("connect");
}

int
main(int argc, char **argv)
{
  char		*t;
  long		len;
  struct stat	st;
  int		fd;

  if (argc<5)
    {
      printf("Usage: %s host port password filename [throttle [blocksize]]\n"
	     "\tCharybdis (this here) assumes that Scylla is at host:port\n"
	     "\tCharybdis authenticates at Scylla with the password.\n"
	     "\tWhen the file is uploaded to Scylla it returns 2 or\n"
	     "\tif Scylla already has the file completely it returns 0.\n"
	     "\tSo run this in a shell loop until true is returned\n"
	     "\tto restart broken uploads seamlessly.\n"
	     "\tThe bandwidth can be controlled with trottle@blocksize\n",
	     argv[0]);
      return 1;
    }

  if (argc>5)
    {
      throttle	= atoi(argv[5]);
      maxblock	= 512;
      if (argc>6)
	maxblock	= atoi(argv[6]);
    }
  if (maxblock<100)
    maxblock	= BUFSIZ;
  else if (maxblock<256)
    maxblock	= 256;

  if ((fd=open(argv[4], O_RDONLY))<0 ||
      fstat(fd, &st) ||
      !S_ISREG(st.st_mode))
    ex("open");

  workmode	= '0';
  alarm_init(1 /* see below */, 300);

  /* Hack: Set a connect timeout to 5 minutes
   */
  alarm(300);
  do_connect(argv[1], atoi(argv[2]));

#ifndef NO_SSL
  alarm_process();
  alarm(300);
  tino_ssl_client();
#endif

  /* Reset timeout to alarm processing
   */
  alarm_reset();
  alarm_process();
  alarm(1 /* see above */);

  workmode	= '1';
  swrite(argv[3]);

  t	= sread();
  if (strncmp(t, "scylla", 6))
    ex("where is my dear scylla?");
  /* leave version control for future
   * we are using version 1.0 protocol
   * (plain password)
   */

  workmode	= '2';
  swrite(argv[4]);

  t	= sread();
  if (!isdigit(*t))
    {
      len	= 0;
      swrite("c");
      workmode	= 'c';
    }
  else
    {
      char	*m;

      sscanf(t, "%ld", &len);
      m	= md5sum(fd, len);
      if (!m)
	ex("sum");

      if (!strcmp(m, t))
	{
	  if (st.st_size==len)
	    {
	      close(fd);
	      /* unlink(argv[4]); */
	      return 0;
	    }
	  swrite("a");
	  workmode	= 'a';
	}
      else
	{
	  len	= 0;
	  swrite("w");
	  workmode	= 'w';
	}
    }
  if (lseek(fd, len, SEEK_SET)!=len)
    ex("lseek");
  totalcount	= st.st_size-len;

  t	= sread();
  if (strcmp("OK", t))
    ex("protocoll failure");

  if (docopy(fd, sock, 0))
    ex("copy error");
  close(fd);

  workmode	= 'e';
  /* Wait for transfer flush */

#ifndef NO_SSL
  tino_ssl_close();
#endif
  close(sock);

  return 2;
}
